(* 横ベクトルの計算 *)
let plus_row u v =
  let dim = Array.length u in
  Array.init dim (fun n -> u.(n) +. v.(n))

let minus_row u v =
  let dim = Array.length u in
  Array.init dim (fun n -> u.(n) -. v.(n))

let prod_row c v =
  let dim = Array.length v in
  Array.init dim (fun n -> c *. v.(n))

(* 行列の計算 *)
let ( +! ) m1 m2 =
  let dim = Array.length m1 in
  Array.init dim (fun n -> plus_row m1.(n) m2.(n))

let ( -! ) m1 m2 =
  let dim = Array.length m2 in
  Array.init dim (fun n -> minus_row m1.(n) m2.(n))

let ( *! ) c m =
  let dim = Array.length m in
  Array.init dim (fun n -> prod_row c m.(n))

let linear_map m v =
  [|
    [| m.(0).(0) *. v.(0).(0) +. m.(0).(1) *. v.(1).(0) |];
    [| m.(1).(0) *. v.(0).(0) +. m.(1).(1) *. v.(1).(0) |]
  |]

(*
 *)
let l_mat p q = 
  [| [| 0.; 1. |]; [| -.q; -2.*.p |] |]

let l_map p q v =
  linear_map (l_mat p q) v

(* 逆行列 *)
let inverse_2 m = 
  let det = m.(0).(0) *. m.(1).(1) -. m.(0).(1) *. m.(0).(1) in
  (1. /. det) *! ([| [| m.(1).(1); -. m.(0).(1) |]; [| -. m.(1).(0); m.(0).(0) |] |])

(* L(p,q)^-1 v *)
let l_mat_inv p q v =
  let l = l_mat p q in
  let l_inv = inverse_2 l in
  linear_map l_inv v

(* exp tL(p,q) v *)
let exp_l_mat p q t v =
  let omega = sqrt (q -. p *. p) in
  let e00 = p *. exp (-.p *. t) /. omega *. sin (omega *. t) +. exp (-.p *. t) *. cos (omega *. t) in
  let e01 = exp (-.p *. t) /. omega *. sin (omega *. t) in
  let e10 = (-. q *. exp (-.p *. t) /. omega) *. sin (omega *. t) in
  let e11 = (-. p *. exp (-.p *. t) /. omega) *. sin (omega *. t) +. exp (-.p *. t) *. cos (omega *. t) in
  let exp_tl = [| [| e00; e01 |]; [| e10; e11 |] |] in
  linear_map exp_tl v
  
(* R(a) v *)
let r_mat a v =
  let r = [| [|0.;0.|]; [|a; 0.|] |] in
  linear_map r v

(*
  semi discrete 法での計算
 *)
let semidiscrete_core lp lq ra y_pre h yn n =
  let ym_ave = 0.5 *! (y_pre.(n) +! y_pre.(n+1)) in
  let yh = exp_l_mat lp lq h yn in
  let yp1 = exp_l_mat lp lq h (l_mat_inv lp lq (r_mat ra ym_ave)) in
  let yp2 = l_mat_inv lp lq (r_mat ra ym_ave) in
  yh +! (0.5 *! (yp1 -! yp2))

(*
    dy/dt(t) = L(p,q)y(t) + R(a)y(t-T) の解の計算
 *)
let solve ~lp ~lq ~ra ~y_pre ~h ~y0 =
  let n_len = Array.length y_pre in
  let rec append_solution y_cur y_pre n =
    let y = semidiscrete_core lp lq ra y_pre h y_cur.(n) n in
    let y_next = Array.append y_cur [| y |] in
    if (n + 2) >= n_len then y_next
    else append_solution y_next y_pre (n+1)
  in
  
  append_solution [| y0 |] y_pre 0

let latest_elem x_arr =
  let n = Array.length x_arr - 1 in
  x_arr.(n)
  

let () =
  let y = Array.make 512 ([| [|0.|]; [|0.|] |]) in
  let lp = 10. in
  let lq = 2000. in
  let ra = 100. in
  let h = 1. /. 2048. in
  let x0 = [| [|1.|]; [|0.|] |] in
  let x = solve ~lp:lp ~lq:lq ~ra:ra ~y_pre:y ~h:h ~y0:x0 in
  Array.iter (fun elm -> print_endline (string_of_float elm.(0).(0))) x;
  let y = x in
  let x0 = latest_elem y in
  let x = solve ~lp:lp ~lq:lq ~ra:ra ~y_pre:y ~h:h ~y0:x0 in
  Array.iter (fun elm -> print_endline (string_of_float elm.(0).(0))) x;
  let y = x in
  let x0 = latest_elem y in
  let x = solve ~lp:lp ~lq:lq ~ra:ra ~y_pre:y ~h:h ~y0:x0 in
  Array.iter (fun elm -> print_endline (string_of_float elm.(0).(0))) x;
  let y = x in
  let x0 = latest_elem y in
  let x = solve ~lp:lp ~lq:lq ~ra:ra ~y_pre:y ~h:h ~y0:x0 in
  Array.iter (fun elm -> print_endline (string_of_float elm.(0).(0))) x
